unit mydatabase;

interface

uses SysUtils, Variants, mylist, myquery;

// typy plikow
type FileType = (XML_FILE,CSV_FILE,UNSUPPORTED);


  {**********************
    TMyTable
  **********************}

type TMyTable = class (TMyList)
  private
    table_name   :string;
    table_file   :string;
    header       :array of string;
    recordscount :integer;
    procedure load_csv(const filename :string);
    procedure load_xml(const filename :string);
    procedure save_csv(const filename :string);
    procedure save_xml(const filename :string);
    function QueryToVariant(const ask :TMyQuery; force:boolean=false) :Variant;
    function compare(const a, b, c :Variant) :boolean;
    function comparenot(const a, b, c :Variant) :boolean;
  public
    fetch_removes:boolean;
    constructor Create; overload;
    constructor Create(const source :TMyTable; const with_content :boolean=false); overload;
    destructor Destroy; override;
    procedure load(const filename :string; ftype :FileType); overload;
    procedure load(const filename :string); overload;
    procedure saveto(const filename :string; ftype :FileType); overload;
    procedure saveto(const filename :string); overload;
    procedure save;
    procedure delete; override;
    function orderby(fieldno :cardinal; ascent :boolean) :boolean; overload;
    function orderby(fieldname :string; ascent :boolean) :boolean; overload;
    function insert(const val :Variant) :Variant; override;
    function fetchrecord: TMyQuery;
    function query(const ask :TMyQuery) :TMyTable;
    property name :string read table_name write table_name;
    property recordsleft: integer read recordscount;
  end;

  {**********************
    BLEDY
  **********************}

type EMyTableError               = class (Exception);
     EMyTableUnsupportedFile     = class (EMyTableError);
     EMyTableNoRecordsLeft       = class (EMyTableError);
     EMyTableQueryUnknownFields  = class (EMyTableError);
     EMyTableCompareError        = class (EMyTableError);
     EMyTableUnknownTable        = class (EMyTableError);

  {**********************
    LANCUCHY ZASOBOWE
  **********************}

resourcestring
  EMYTABLEUNSUPPORTEDFILE_MSG = 'TMyTable: Unsupported file';
  EMYTABLENORECORDSLEFT_MSG   = 'TMyTable: No records left';
  EMYTABLEUNKNOWNFIELDS_MSG   = 'TMyTable: Unknown fields in query';
  EMYTABLEUNKNOWNTABLE_MSG    = 'TMyTable: Unknown table name';
  EMYTABLE_UNCOMPARABLE_MSG   = 'TMyTable: Uncomparable fields';

implementation


  {**********************
    TMyTable
  **********************}

// utworzenie pustej tabeli
constructor TMyTable.Create;
begin
inherited Create;
recordscount:=0;
fetch_removes:=false;
table_name := 'table1';
end;

// utworzenie tabeli z kolumnami i nazwa jak source
// jesli with_content=true to rowniez kopiuje dane
constructor TMyTable.Create(const source :TMyTable; const with_content :boolean=false);
var i:integer;
begin
if source<>nil then
  begin
  inherited Create;
  self.recordscount:=0;
  self.table_name := source.table_name;
  SetLength(self.header, High(source.header)+1);
  for i:=0 to High(source.header) do
    self.header[i]:=source.header[i];
  self.fetch_removes:=false;
  if (with_content=true) and  (not source.empty) then
    begin
    source.fetch_removes:=false;
    for i:=0 to source.recordscount-1 do
      begin
      insert(source.current);
      source.next;
      end;
    next; // przejdz z ostatniego dodanego, na pierwszy element
    end;
  end;
end;

destructor TMyTable.Destroy;
begin
inherited Destroy;
end;

// usuniecie elementu z tabeli i obnizenie licznika
procedure TMyTable.delete;
begin
dec(recordscount);
inherited delete;
end;

// dodanie elementu do tabeli i zwiekszenie licznika
function TMyTable.insert(const val :Variant) :Variant;
begin
inc(recordscount);
inherited insert(val);
end;

// odczytanie aktualnego rekordu i przejscie do nastepnego
// jesli fetch_removes = true to takze usuwa przeczytany rekord
// jesli licznik wskazuje 0 elementow, podnosi EMyTableNoRecordsLeft
function TMyTable.fetchrecord: TMyQuery;
var i        :integer;
    rslt     :TMyQuery;
    tmpfield :TFieldItem;
begin
if (self.recordsleft=0) or (empty) then
  raise EMyTableNoRecordsLeft.Create(EMYTABLENORECORDSLEFT_MSG);
rslt := TMyQuery.Create;
rslt.operation := QINSERT;
rslt.table := table_name;
for i:=0 to High(header) do
  begin
  tmpfield.name:=header[i];
  tmpfield.value:=getcurrent[i];
  rslt.field_add(tmpfield);
  end;
if fetch_removes=true then
  delete
else next;
Result := rslt;
end;

// zaladowanie danych z pliku, scisle okreslenie rozszezenia
procedure TMyTable.load(const filename: string; ftype :FileType);
begin
case ftype of
  XML_FILE :load_xml(filename);
  CSV_FILE :load_csv(filename);
end;
table_file:=filename;
table_name:=copy(filename, LastDelimiter('\', filename)+1, Length(filename)-LastDelimiter('\', filename)-4);
end;

// zaladowanie danych z pliku
procedure TMyTable.load(const filename: string);
var filename_ext :string;
begin
filename_ext := copy(filename, length(filename)-2, 3);
if filename_ext='xml' then load(filename, XML_FILE)
else if filename_ext='csv' then load(filename, CSV_FILE)
else raise EMyTableUnsupportedFile.Create(EMYTABLEUNSUPPORTEDFILE_MSG);
end;

// zapisanie danych do pliku, scisle okreslenie rozszezenia
procedure TMyTable.save;
begin
saveto(table_file);
end;

// zapisanie danych do pliku, scisle okreslenie rozszezenia
procedure TMyTable.saveto(const filename :string; ftype :FileType);
begin
case ftype of
  XML_FILE :save_xml(filename);
  CSV_FILE :save_csv(filename);
end;
end;

// zapisanie danych do pliku
procedure TMyTable.saveto(const filename: string);
var filename_ext :string;
begin
filename_ext := copy(filename, length(filename)-2, 3);
if filename_ext='xml' then saveto(filename, XML_FILE)
else if filename_ext='csv' then saveto(filename, CSV_FILE)
else raise EMyTableUnsupportedFile.Create(EMYTABLEUNSUPPORTEDFILE_MSG);
end;


// zaladowanie danych z pliku csv
procedure TMyTable.load_csv(const filename :string);
const CSV_DELIM = ';';
var i,c    :integer;
    t      :Text;
    buf    :string;
    recbuf :array of string;
begin
try
  AssignFile(t, filename);
  Reset(t);
  recordscount:=0;
  // czyta wiersz naglowkow tabeli
  readln(t, buf);
  i:=pos(CSV_DELIM, buf);
  while (i<>0) or (length(buf)>0) do
    begin
    SetLength(header, High(header)+2);
    if i=0 then i:=length(buf);
    header[High(header)] :=copy(buf, 0, i-1);
    buf := copy(buf, i+1, length(buf)-i);
    i:=pos(CSV_DELIM, buf);
    end;
  // czyta wiersze tabeli
  SetLength(recbuf, High(header)+1);
  while not eof(t) do
    begin
    readln(t, buf);
    for c:=0 to High(header) do
      begin
      i:=pos(CSV_DELIM, buf);
      recbuf[c] := copy(buf, 0, i-1);
      buf := copy(buf, i+1, length(buf)-i);
      end;
    insert(recbuf);
    end;
finally
  Close(t);
end;
end;

// zapisanie danych do pliku csv
procedure TMyTable.save_csv(const filename :string);
const CSV_DELIM = ';';
var i      :integer;
    t      :Text;
    buf    :TMyQuery;
    tmpbuf :Variant;
begin
try
  AssignFile(t, filename);
  Rewrite(t);
  // zapisuje naglowek
  for i:=0 to High(header) do
    write(t, header[i]+CSV_DELIM);
  writeln(t);
  // zapisuje wiersze tabeli
  fetch_removes:=true;
  buf :=TMyQuery.Create;
  while recordscount>0 do
    begin
    buf:=self.fetchrecord;
    tmpbuf:=self.QueryToVariant(buf);
    for i:=0 to High(header) do
      write(t, tmpbuf[i]+CSV_DELIM);
    writeln(t);
    end;
  buf.Destroy;
finally
  Close(t);
end;
end;

// zaladowanie danych z pliku xml
procedure TMyTable.load_xml(const filename: string);
begin
end;

// zapisanie danych do pliku xml
procedure TMyTable.save_xml(const filename :string);
begin
end;

// jesli dane pola pasuja do bazy, zwraca postac wariantowa gotowa
// do dodania, lub update
// w przeciwnym wypadku zwraca false
// funckja wywolana z Query z polem * zwroci liste pol w tabeli
// jesli w ask sa pola, ktorych nie ma w tabeli result:= false
// ale jesli jest force=true, to result zwraca variant tylko z elementami
// ktore sa w tabeli
function TMyTable.QueryToVariant(const ask :TMyQuery; force:boolean=false) :Variant;
var i, j  :integer;
    adbuf :array of string;
    b     :boolean;
begin
setlength(adbuf, High(header)+1);
for i:=0 to High(header) do
  adbuf[i]:='';
If ask.table<>self.table_name then
  Raise EMyTableUnknownTable.Create(EMYTABLEUNKNOWNTABLE_MSG)
else if ask.fields=nil then b:=false
else if ask.fields[0].name='*' then
  begin
  for i:=0 to High(header) do
    adbuf[i]:=header[i];
  b:=true;
  end
else
  begin
  i:=0;
  b:=true;
  while (i<=High(ask.fields)) and (b<>false) do
    begin
    b:=false;
    for j:=0 to High(header) do
      if ask.fields[i].name=header[j] then
        begin
        b:=true;
        adbuf[j]:=ask.fields[i].value;
        end;
    if force then b:=true;
    inc(i);
    end;
  end;
if b=true then
  Result:=adbuf
else Result:=b;
end;

function TMyTable.compare(const a, b, c :Variant) :boolean;
begin
if ((VarType(a) and VarArray)=VarArray) and ((VarType(b) and VarArray)=VarArray) then
  begin
  if CompareStr(UpperCase(a[c]), UpperCase(b[c]))<0 then
    Result:=true
  else Result:=false;
  end
else Raise EMyTableCompareError.Create(EMYTABLE_UNCOMPARABLE_MSG);
end;

function TMyTable.comparenot(const a, b, c :Variant) :boolean;
begin
Result :=not compare(a, b, c);
end;

// sortuje tabele wg pola o numerze fieldno
function TMyTable.orderby(fieldno :cardinal; ascent :boolean) :boolean;
begin
if fieldno>High(header) then
  Raise EMyTableQueryUnknownFields.Create(EMYTABLEUNKNOWNFIELDS_MSG+' (ORDER BY FIELDNO)')
else
  begin
  if ascent then
    sort(compare, fieldno)
  else sort(comparenot, fieldno);
  Result:=true;
  end;
end;

// sortuje tabele wg pola o nazwie fieldname
// jesli jest takie pole, zwraca rezultat dzialania orderby(cardinal)
// jesli nie ma takiego pola, zwraca false
function TMyTable.orderby(fieldname :string; ascent :boolean) :boolean;
var i :integer;
begin
i:=0;
while (i<=High(header)) and (fieldname<>header[i]) do
  inc(i);
if (i<=High(header)) and (fieldname=header[i]) then
    Result :=orderby(i, ascent)
else Raise EMyTableQueryUnknownFields.Create(EMYTABLEUNKNOWNFIELDS_MSG+' (ORDER BY FIELDNAME)');
end;

// przetworzenie zapytania. jesli zapytanie typu select, zwraca
// tabele wynikowa, w pozostalych przypadkach zwraca nil
// jesli zapytanie ma inne pola niz tabela, podnosi EMyTableUnknownFields
// jesli zapytanie odnosi sie do tabeli o innej nazwie, podnosi
// EMyTableUnknownTable
function TMyTable.query(const ask: TMyQuery): TMyTable;
var rslt           :TMyTable;
    bufquery       :TMyQuery;
    i, j           :integer;
    tmp, updatetmp :Variant;
begin
bufquery := TMyQuery.Create;
fetch_removes:=false;
Result := nil;
// czy pola pasuja do bazy
if (VarType(QueryToVariant(ask))=vtBoolean) then
  Raise EMyTableQueryUnknownFields.Create(EMYTABLEUNKNOWNFIELDS_MSG);
if ask.operation=QSELECT then // SELECT
  begin
  // przygotuj tabele wynikowa
  if ask.fields[0].name='*' then
    rslt := TMyTable.Create(self) // na obraz i podobienstwo selfa, ale bez danych
  else
    begin
    rslt := TMyTable.Create;
    rslt.name := name;
    SetLength(rslt.header, High(ask.fields)+1);
    for i:=0 to High(ask.fields) do
      rslt.header[i]:=ask.fields[i].name;
    end;
  // przetworz dane i dodaj do tabeli wynikowej
  for i:=0 to recordscount-1 do
    begin
    bufquery:=fetchrecord;
    if bufquery.conditions_aply(ask) then
      begin
      tmp:=rslt.QueryToVariant(bufquery, true);
      rslt.insert(tmp);
      end;
    end;
  // wykonaj sortowanie, jesli trzeba
  if ask.order<>'' then
    rslt.orderby(ask.order, ask.order_asc);
  Result := TMyTable.Create(rslt, true);
  end
else if ask.operation=QUPDATE then // UPDATE
  begin
  updatetmp:=QueryToVariant(ask);
  for i:=0 to recordscount-1 do
    begin
    bufquery:=fetchrecord;
    if bufquery.conditions_aply(ask) then
      begin
      tmp:=QueryToVariant(bufquery, true);
      for j:=0 to High(header) do
        if updatetmp[j]<>'' then tmp[j]:=updatetmp[j];
      prev;
      delete;
      prev;
      insert(tmp);
      next;
      end;
    end;
  end
else if ask.operation=QINSERT then // INSERT
  begin
  tmp:=QueryToVariant(ask);
  insert(tmp);
  end
else if ask.operation=QDELETE then // DELETE
  for i:=0 to recordscount-1 do
    begin
    bufquery:=fetchrecord;
    if bufquery.conditions_aply(ask) then
      begin
      prev;
      delete;
      end;
    end;
fetch_removes:=true;
bufquery.Destroy;
end;
end.
