unit mylist;

interface

uses SysUtils, Variants;

  {**********************
    TMyListItem
  **********************}

type
  PTMyListItem = ^TMyListItem;
  TMyListItem = class (TObject)
    private
      data  :Variant;
      instance  :pointer;
      pnext :PTMyListItem;
      pprev :PTMyListItem;
    public
      constructor Create(instance :pointer);
      property value :Variant read data;
      property next :PTMyListItem read pnext;
      property prev :PTMyListItem read pprev;
  end;

  {**********************
    TMyList
  **********************}
  TSortfuncob =function(const a, b, c:Variant) :boolean of object;

  TMyList = class (TObject)
    protected
      pitems :PTMyListItem;
      function getcurrent :Variant;
      function switchnext :Variant;
      function switchprev :Variant;
    public
      constructor Create;
      destructor Destroy; override;
      property current :Variant read getcurrent;
      property next :Variant    read switchnext;
      property prev :Variant    read switchprev;
      function insert(const val :Variant) :Variant; virtual;
      procedure delete; virtual;
      procedure ReplaceWithNext;
      procedure sort(sortfunc :TSortfuncob; constant :Variant);
      function empty :boolean;
      function count :integer;
      function items(i :integer) :variant;
  end;


  {**********************
    TYPY BLEDOW
  **********************}
type EMyListError      = class (Exception);
     EListNoItemsLeft  = class (EMyListError);
     EListOutOfMemory  = class (EMyListError);
     EListNoItemsFound = class (EMyListError);


  {**********************
    LANCUCHY ZASOBOWE
  **********************}

resourcestring ELISTNOITEMSLEFT_MSG  = 'TMyList: No items left to delete.';
               ELISTOUTOFMEMORY_MSG  = 'TMyList: Not enought memory to insert item.';
               ELISTNOITEMSFOUND_MSG = 'TMyList: No items found.';

  {**********************
    IMPLEMENTATION
  **********************}

implementation


  {**********************
    TMyList
  **********************}

// zwraca wartosc aktualnego elementu
// gdy brak elementow podnosi EListNoItemsFound
function TMyList.getcurrent :Variant;
begin
if pitems=nil then
  raise EListNoItemsFound.Create(ELISTNOITEMSFOUND_MSG)
else
  Result := pitems^.data;
end;

// przechodzi do poprzedniego elementu listy i zwraca jego wartosc
// gdy brak elementow podnosi EListNoItemsFound
function TMyList.switchnext :Variant;
begin
if (pitems=nil) or (pitems^.pnext=nil) then
  raise EListNoItemsFound.Create(ELISTNOITEMSFOUND_MSG)
else
  begin
  pitems:=pitems.pnext;
  Result := getcurrent;
  end;
end;

// przechodzi do nastepnego elementu listy i zwraca jego wartosc
// gdy brak elementow podnosi EListNoItemsFound
function TMyList.switchprev :Variant;
begin
if (pitems=nil) or (pitems^.pprev=nil) then
  raise EListNoItemsFound.Create(ELISTNOITEMSFOUND_MSG)
else
  begin
  pitems:=pitems.pprev;
  Result := getcurrent;
  end;
end;

// Dodaje element do listy i przechodzi do niego,
// zwraca wartosc dodanego elementu
// podnosi EListOutOfMemory, gdy zabraknie pamieci
function TMyList.insert(const val :Variant) :Variant;
var tmp :PTMyListItem;
begin
try
  if pitems=nil then
    begin
    New(pitems);
    pitems^ := TMyListItem.Create(pitems);
    pitems^.data:=val;
    end
  else
    begin
    New(tmp);
    TMyListItem(tmp^) :=TMyListItem.Create(tmp);
    tmp^.data :=val;
    tmp^.pnext :=pitems^.pnext;
    tmp^.pprev :=pitems;
    pitems^.pnext^.pprev:=tmp;
    pitems^.pnext:=tmp;
    end;
  Result:=switchnext;
except
  on EOutOfMemory do
    raise EListOutOfMemory.Create(ELISTOUTOFMEMORY_MSG);
end;
end;

// Usuwa element listy i przechodzi do nastepnego
// jesli ostatni element listy - lista=nil (stan poczatkowy)
// podnosci EListNoItemsLeft przy probie usuniecia ostatniego elementu
procedure TMyList.delete;
var ptmp :PTMyListItem;
begin
if pitems=nil then
  raise EListNoItemsLeft.Create(ELISTNOITEMSLEFT_MSG)
else
  begin
  if pitems^.instance=pitems^.pnext then // usuwa ostatni element listy
    begin
    pitems.Destroy;
    pitems:=nil;
    end
  else // usuwa szeregowy element listy
    begin
    ptmp:=pitems^.pprev;
    switchnext;
    pitems^.pprev.Destroy;
    pitems^.pprev:=ptmp;
    pitems^.pprev^.pnext:=pitems^.instance;
    end;
  end;
end;

function TMyList.empty :boolean;
begin
if pitems=nil then Result:=true
else Result:=false;
end;

function TMyList.count :integer;
var p :pointer;
begin
Result:=0;
p:=pitems;
repeat
  switchnext;
  inc(Result);
until p=pitems;
end;

procedure TMyList.ReplaceWithNext;
var buf :Variant;
begin
if pitems=nil then
  raise EListNoItemsLeft.Create(ELISTNOITEMSLEFT_MSG)
else
  begin
  buf := pitems^.next^.data;
  pitems^.next^.data := pitems^.data;
  pitems^.data := buf;
  end;
end;

procedure TMyList.sort(sortfunc :TSortfuncob; constant :Variant);
var i, j :integer;
begin
if (not empty) and (count>=2) then
  begin
  for i:=0 to count-1 do
    begin
    for j:=0 to count-2 do
      begin
      if sortfunc(pitems^.data, pitems^.next^.data, constant)=false then
        ReplaceWithNext;
      next;
      end;
    next;
    end;
  end;
end;

function TMyList.items(i :integer) :variant;
var n :integer;
begin
while (i>count) or (-i<count) do
  if i>0 then i:=i-count else
              i:=i+count;
if i>0 then for n:=0 to i do next
       else for n:=0 to i do prev;
Result:=current;
if i>0 then for n:=0 to i do prev
       else for n:=0 to i do next;
end;

constructor TMyList.Create;
begin
inherited Create;
pitems := nil;
end;

destructor TMyList.Destroy;
begin
if pitems<>nil then
  Dispose(pitems); // zwalnia pamiec zaalokowana przez Constructor
inherited Destroy;
end;

  {**********************
    TMyListItem
  **********************}

constructor TMyListItem.Create(instance :pointer);
begin
inherited Create;
self.instance:=instance;
self.pprev:=instance;
self.pnext:=instance;
end;

end.
