unit myquery;

interface

uses SysUtils, StrUtils;

  {**********************
    TMyQuery
  **********************}

type
  // typy skladowe TMyQuery
  TQueryOperationType = (QSELECT, QINSERT, QDELETE, QUPDATE);
  TFieldItem  = record
    name, value :string;
  end;
  TConditionItem  = record
    name, value :string;
  end;
  TQueryFieldsList     = array of TFieldItem;
  TQueryConditionsList = array of TConditionItem;

  // TMYQUERY
  TMyQuery= class
    protected
      table_name      :string;
      orderby         :string;
      ascent          :boolean;
      operation_type  :TQueryOperationType;
      fields_list     :TQueryFieldsList;
      conditions_list :TQueryConditionsList;
      function sql_string :string;
      procedure parse_sql(s :string);
      procedure setconditions(conditions :TQueryConditionsList);
      procedure setfields(fields :TQueryFieldsList);
      function SafePos(Substr, S: string): Integer;
    public
      destructor Destroy; override;
      procedure field_add(new :TFieldItem); overload;
      procedure field_add(name, value :string); overload;
      procedure condition_add(new :TConditionItem); overload;
      procedure condition_add(name, value :string); overload;
      function conditions_aply(ask :TMyQuery) :boolean;
      property operation  :TQueryOperationType  read operation_type  write operation_type;
      property table      :string               read table_name      write table_name;
      property order      :string               read orderby         write orderby;
      property order_asc  :boolean              read ascent          write ascent;
      property fields     :TQueryFieldsList     read fields_list     write setfields;
      property conditions :TQueryConditionsList read conditions_list write setconditions;
      property sql        :string               read sql_string      write parse_sql;
  end;

type
     EMyQueryError               = class (Exception);
     EMyQueryParseError          = class (EMyQueryError);

resourcestring
  EMYQUERYERROR_MSG                       = 'TMyQuery error';
  EMYQUERYPARSEERROR_MSG                  = 'TMyQuery parse error';
  EMYQUERYPARSEERROR_UNKNOWN_OPERTYPE_MSG = 'TMyQuery parser: unknown operation type';
  EMYQUERYPARSEERROR_UNEXPECTED_MSG       = 'TMyQuery parser unexpected:';
  EMYQUERYPARSEERROR_EXPECTED_MSG         = 'TMyQuery parser expected:';


implementation

  {**********************
    TMyQuery
  **********************}

// zwalnia pamiec przeznaczona na liste pol i warunkow
destructor TMyQuery.Destroy;
begin
Finalize(fields_list);
Finalize(conditions_list);
end;

// alokuje pamiec i ustawia liste pol (fields)
procedure TMyQuery.setfields(fields :TQueryFieldsList);
var i :integer;
begin
for i:=0 to High(fields) do
  field_add(fields[i]);
end;

function TMyQuery.SafePos(Substr, S: string): Integer;
var rslt, cite_start, cite_end :integer;
begin
rslt := pos(substr, s);
cite_start := pos('"', s);
cite_end := posEx('"', s, cite_start+1);
if (cite_end>cite_start) and (rslt in [cite_start..cite_end]) then
  rslt := PosEx(substr, s, cite_end);
Result := rslt;
end;

// alokuje pamiec i ustawia liste warunkow (conditions)
procedure TMyQuery.setconditions(conditions :TQueryConditionsList);
var i :integer;
begin
for i:=0 to High(conditions) do
  condition_add(conditions[i]);
end;

// dodaje pole do zapytania
procedure TMyQuery.field_add(new :TFieldItem);
begin
setlength(fields_list, high(fields_list)+2);
fields_list[high(fields_list)] := new;
end;

procedure TMyQuery.field_add(name, value :string);
var new :TFieldItem;
begin
new.name:=name;
new.value:=value;
field_add(new);
end;

// dodaje warunek do zapytania
procedure TMyQuery.condition_add(new :TConditionItem);
begin
setlength(conditions_list, high(conditions_list)+2);
conditions_list[high(conditions_list)]:=new;
end;

procedure TMyQuery.condition_add(name, value :string);
var new :TConditionItem;
begin
new.name:=name;
new.value:=value;
condition_add(new);
end;

// PRAWDA, jesli self spelnia warunki z zapytania w argumencie
function TMyQuery.conditions_aply(ask :TMyQuery) :boolean;
var i, j :integer;
    b    :boolean;
begin
i:=0;
b:=true;
if ask.conditions=nil then b:=true
else while (i<=High(ask.conditions)) and (b=true) do
  begin
  for j:=0 to High(fields_list) do
    begin
    if (ask.conditions[i].name=fields_list[j].name) and (ask.conditions_list[i].value<>fields_list[j].value) then
      b:=false;
    end;
  if (j=High(fields_list)) and (ask.conditions[i].name=fields_list[j].name) then
    b:=false;
  inc(i);
  end;
Result:=b;
end;

// zwraca Query w postaci lancucha
// jesli jest Query typu SELECT, INSERT, lub UPDATE i ma pusta liste pol
// podnosi EMyQueryError
function TMyQuery.sql_string :string;
var buf :string;
    i   :integer;
begin
case operation_type of
  QSELECT:
    begin
    buf:='SELECT ';
    if fields_list=nil then
      raise EMyQueryError.Create(EMYQUERYERROR_MSG);
    for i:=0 to High(fields_list) do
      buf:=buf+fields_list[i].name+', ';
    buf:=trim(buf);
    buf[length(buf)]:=' ';
    buf:=buf+'FROM "'+table_name+'" ';
    end;
  QINSERT: buf:='INSERT "'+table_name+'" SET ';
  QUPDATE: buf:='UPDATE "'+table_name+'" SET ';
  QDELETE: buf:='DELETE FROM "'+table_name+'" ';
end;
if operation_type in [QINSERT, QUPDATE] then
  begin
    if fields_list=nil then
      raise EMyQueryError.Create(EMYQUERYERROR_MSG);
    for i:=0 to High(fields_list) do
      buf:=buf+fields_list[i].name+'="'+fields_list[i].value+'", ';
    buf[length(buf)-1]:=' ';
  end;
if not (operation_type=QINSERT) and (conditions_list<>nil) then
  begin
    buf:=buf+'WHERE ';
    for i:=0 to High(conditions_list) do
      buf:=buf+conditions_list[i].name+'="'+conditions_list[i].value+'" AND ';
    buf:=copy(buf, 0, length(buf)-5);
  end;
if (operation_type=QSELECT) and (orderby<>'') then
  begin
  buf:=buf+' ORDER BY '+orderby;
  if not ascent then buf:=buf+' DESC';
  end;
Result := trim(buf);
end;

// informacje zawarte w lancuchu s wprowadza do pol objektu
procedure TMyQuery.parse_sql(s :string);
var buf :string;
    i   :integer;
begin
// typ operacji (SELECT/itp...)
s :=trimleft(s);
buf := copy(s, 0, pos(' ', s)-1);
if buf='SELECT' then operation_type:=QSELECT
else if buf='INSERT' then operation_type:=QINSERT
else if buf='UPDATE' then operation_type:=QUPDATE
else if buf='DELETE' then operation_type:=QDELETE
else raise EMyQueryParseError.Create(EMYQUERYPARSEERROR_UNKNOWN_OPERTYPE_MSG);
s :=trimleft(copy(s, pos(' ', s), length(s)));

// nazwy pol w SELECT ... FROM
if operation_type=QSELECT then
  begin
  buf := copy(s, 0, pos(' ', s)-1);
  s :=trimleft(copy(s, pos(' ', s), length(s)));
  i:=1;
  if buf='*' then
    begin
    SetLength(fields_list, 1);
    fields_list[0].name:='*';
    s := trimleft(copy(s, pos('FROM', s)+4, length(s)));
    end
  else if buf='FROM' then raise EMyQueryParseError.Create(EMYQUERYPARSEERROR_UNEXPECTED_MSG+buf)
  else
    while (buf<>'FROM') do
      begin
        if pos('FROM', s)=0 then raise EMyQueryParseError.Create(EMYQUERYPARSEERROR_EXPECTED_MSG+'"FROM"');
        setlength(fields_list, i);
        with fields_list[i-1] do
          begin
          name := trim(copy(buf, 0, pos(',', buf)-1));
          if name='' then
            name :=trim(buf);
          end;
      buf := copy(s, 0, pos(' ', s)-1);
      s :=trimleft(copy(s, pos(' ', s), length(s)));
      inc(i);
      end;
  end;

// rozpoznanie FROM w DELETE FROM
if operation_type=QDELETE then
  begin
  buf := copy(s, 0, pos(' ', s)-1);
  if buf<>'FROM' then
    raise EMyQueryParseError.Create(EMYQUERYPARSEERROR_UNEXPECTED_MSG+buf);
  s :=trimleft(copy(s, pos(' ', s), length(s)));
  end;

// nazwa tabeli
s:=copy(s, 2, length(s)); // omin pierwszy cudzyslow w nazwie tabeli
table_name := copy(s, 0, pos('"', s)-1);
s :=trimleft(copy(s, pos('"', s)+1, length(s)));

// SET (tylko INSERT i UPDATE),
buf := copy(s, 0, pos(' ', s)-1);
s :=trimleft(copy(s, pos(' ', s), length(s)));

if (buf='SET') then
  begin
  if not (operation_type in [QINSERT, QUPDATE]) then
    raise EMyQueryParseError.Create(EMYQUERYPARSEERROR_UNEXPECTED_MSG+buf);

  buf := copy(s, 0, safepos(' ', s)-1);
  if buf='' then
    begin
    buf:=s;
    s:='';
    end
  else
    s :=trimleft(copy(s, safepos(' ', s), length(s)));
  i:=1;
  while (buf<>'WHERE') and (length(buf)>0) do
    begin
      setlength(fields_list, i);
      with fields_list[i-1] do
        begin
        name := trim(copy(buf, 0, pos('=', buf)-1));
        buf :=trimleft(copy(buf, pos('=', buf)+1, length(buf)));
        value := copy(buf, 0, safepos(',', buf)-1);
        if value='' then
          value:=buf;
        value:=copy(value, 2, length(value)-2);
        inc(i);
        end;
    buf := copy(s, 0, safepos(' ', s)-1);
    if buf='' then
      begin
      buf:=s;
      s:='';
      end
    else
      s :=trimleft(copy(s, safepos(' ', s), length(s)));
    end;
{
  buf := copy(s, 0, pos(' ', s)-1);
  if buf='' then
    begin
    buf:=s;
    s:='';
    end
  else
    s :=trimleft(copy(s, pos(' ', s), length(s)));
  i:=1;
  while (buf<>'WHERE') and (length(buf)>0) do
    begin
      setlength(fields_list, i);
      with fields_list[i-1] do
        begin
        name := trim(copy(buf, 0, pos('=', buf)-1));
        buf :=trimleft(copy(buf, pos('=', buf)+1, length(buf)));
        value := copy(buf, 0, pos(',', buf)-1);
        if value='' then
          value:=buf;
        value:=copy(value, 2, length(value)-2);
        inc(i);
        end;
    buf := copy(s, 0, pos(' ', s)-1);
    if buf='' then
      begin
      buf:=s;
      s:='';
      end
    else
      s :=trimleft(copy(s, pos(' ', s), length(s)));
    end;
}
  end;

// WHERE
if (buf='WHERE') then
  begin
  if operation_type=QINSERT then
    raise EMyQueryParseError.Create(EMYQUERYPARSEERROR_UNEXPECTED_MSG+buf);
  buf := copy(s, 0, safepos(' ', s)-1);
  if buf='' then
    begin
    buf:=s;
    s:='';
    end
  else
    s :=trimleft(copy(s, safepos(' ', s), length(s)));
  i:=1;
  while (buf<>'ORDER') and (length(buf)>0) do
    begin
      if buf<>'AND' then
        begin
        setlength(conditions_list, i);
        with conditions_list[i-1] do
          begin
          name := trim(copy(buf, 0, pos('=', buf)-1));
          buf :=trimleft(copy(buf, pos('=', buf)+1, length(buf)));
          value:=copy(buf, 2, length(buf)-2);
          end;
        inc(i);
        end;
    buf := copy(s, 0, safepos(' ', s)-1);
    if buf='' then
      begin
      buf:=s;
      s:='';
      end
    else
      s :=trimleft(copy(s, safepos(' ', s), length(s)));
    end;
  end;

// ORDER BY
if buf='ORDER' then
  begin
  buf := copy(s, 0, pos(' ', s)-1);
  s :=trimleft(copy(s, pos(' ', s), length(s)));
  if buf<>'BY' then
    raise EMyQueryParseError.Create(EMYQUERYPARSEERROR_EXPECTED_MSG+'"BY"');
  buf := copy(s, 0, pos(' ', s)-1);
  if buf='' then
    begin
    buf:=s;
    s:='';
    ascent:=true; // jesli to koniec zapytania, domyslnie ustaw ascent
    end
  else
    begin
    s :=trimleft(copy(s, pos(' ', s), length(s)));
    if pos('DESC', s)>0 then ascent:=false
    else ascent:=true;
    end;
  orderby:=buf;
  end;

end;

end.
 