unit service_main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, SvcMgr, Dialogs, Registry, ScktComp,
  myquery,
  mydatabase,
  Sockets;

type
  TDelphiapp2 = class(TService)
    TcpServer1: TTcpServer;
    procedure TcpServer1Accept(Sender: TObject; ClientSocket: TCustomIpClient);
    procedure ServiceExecute(Sender: TService);
    procedure ServiceStart(Sender: TService; var Started: Boolean);
    procedure ServiceStop(Sender: TService; var Stopped: Boolean);
    procedure ServiceCreate(Sender: TObject);
    procedure ServiceDestroy(Sender: TObject);
  private
    { Private declarations }
  public
    tbl          :array of TMyTable;  // tabela bazy danych
    starttime    :TDateTime; // data wystartowania serwisu
    echo_queries :boolean;   // czy zwracac zapytania
    log_queries  :boolean;   // czy zwracac zapytania
    logqueryfile :string;    // nazwa pliku, w ktorym logowac
    function GetServiceController: TServiceController; override;
    procedure LogQuery(query :string);
  end;

// stale wykorzystywane przez LogMessage()
const
// LOGMESSAGE: CATEGORY
      INFO    = 0; // category
// LOGMESSAGE: ID
      ID_UP      = 1; // service start
      ID_DOWN    = 2; // service stop
      ID_DLLS    = 3; // dll
      ID_TBL     = 4; // table
      ID_TCP     = 5; // tcp
      ID_CREATE  = 6; // ServiceCreate
      ID_TBLLOAD = 7; // table load
      ID_IERR    = 8; // internal error
      ID_UERR    = 9; // unexpected error

type EStartupError = class (Exception);
     ENoSuchTable  = class (Exception);

resourcestring
   APP_VERSION = 'Ver.1'; // numer wersji aplikacji

   // lancuchy zapisywane w rejestrze zdarzen
   LOG_SERVICE_UP     = 'Service is up';
   LOG_SERVICE_DOWN   = 'Service is down';
   LOG_ERROR          = 'Error: ';
   LOG_CLIENT         = 'Client connection';
   TABLELOAD_MSG      = 'Loading table: ';
   ESTARTUPERROR_MSG  = 'Startup error: ';
   ENOSUCHTABLE_MSG   = 'No such table: ';
   EINTERNALERROR_MSG = 'Internal error';

   // standardowe wiadomosci
   STDMSG_GOODBYE = 'Goodbye';
   STDMSG_HELLO   = 'Hello ';
   STDMSG_YES     = 'YES';
   STDMSG_NO      = 'NO';
   STDMSG_CONFIRM = 'OK'; // wysylane, jesli zapytanie inne niz SELECT
   STDMSG_ERR     = 'ERROR '; // tak standardowa! zeby klient wiedzial, ze to nie dane

   // polecenia rozumiane przez program, poza zapytaniami
   STDQUESTION_UP   = 'UPTIME';
   STDQUESTION_TBLN = 'TABLES';
   STDQUESTION_VER  = 'VERSION';
   STDQUESTION_ECHO = 'ECHO';


var
  Delphiapp2     :TDelphiapp2; // instancja programu
  id             :HMODULE;   // zasob z DLL do obslugi rejestrow
  ConfigGetValue :function(const option: PChar) :PChar;

implementation

{$R *.DFM}

procedure ServiceController(CtrlCode: DWord); stdcall;
begin
  Delphiapp2.Controller(CtrlCode);
end;

function TDelphiapp2.GetServiceController: TServiceController;
begin
  Result := ServiceController;
end;

// dodaje na koniec logu ciag query z data wykonania
procedure TDelphiapp2.LogQuery(query :string);
var t:Text;
begin
try
  AssignFile(t, logqueryfile);
  try
  Append(t);
  except
    on EInOutError do
     Rewrite(t);
  end;
  Writeln(t, DateTimeToStr(Now)+#9+query);
finally
  Close(t);
end;
end;

// zapisanie informacji o polaczeniu w dzienniku systemowym
// komunikacja z klientem
procedure TDelphiapp2.TcpServer1Accept(Sender: TObject; ClientSocket: TCustomIpClient);
var s        :string;
    i        :integer;
    query    :TMyQuery;
    outtable :TMyTable;
begin
LogMessage(LOG_CLIENT+' '+ClientSocket.RemoteHost+':'+ClientSocket.RemotePort, EVENTLOG_INFORMATION_TYPE, INFO, ID_TCP);
ClientSocket.Sendln(STDMSG_HELLO+ClientSocket.RemoteHost);
repeat
  try // komunikat bledu trafi do klienta i zakonczy polaczenie
  s := ClientSocket.Receiveln;
  // czy lancuch jest ktoryms z polecen rozumianych przez baze
  if s<>'' then
  if s=STDQUESTION_VER then ClientSocket.Sendln(APP_VERSION)
  else if s=STDQUESTION_UP then ClientSocket.Sendln(IntToStr(round((now-starttime)*24*60))+' minutes')
  else if s=STDQUESTION_ECHO then
    begin
    if echo_queries then ClientSocket.Sendln(STDMSG_YES)
                    else ClientSocket.Sendln(STDMSG_NO);
    end
  else if s=STDQUESTION_TBLN then
    begin
      ClientSocket.Sendln(IntToStr(High(tbl)+1));
      for i:=0 to High(tbl) do
        ClientSocket.Sendln(tbl[i].name)
    end
  else // lancuch traktujemy jako zapytanie
    begin
    // parsuj query
    query := TMyQuery.Create;
    query.sql:=s;
    // okresl, do ktorej bazy
    i:=0;
    while (i<=High(tbl)) and (tbl[i].name<>query.table) do inc(i);
    if (i>High(tbl)) or (tbl[i].name<>query.table) then
      Raise ENoSuchTable.Create(ENOSUCHTABLE_MSG+query.table);
     // zwroc dane
    if echo_queries then ClientSocket.Sendln(query.sql);
    if log_queries then LogQuery(query.sql);
    if query.operation<>QSELECT then
      begin
      tbl[i].query(query);
      ClientSocket.Sendln(STDMSG_CONFIRM);
      end
    else
      begin
      outtable:=TMyTable.Create(tbl[i].query(query), true);
      if outtable<>nil then
        begin
        ClientSocket.Sendln(STDMSG_CONFIRM);
        ClientSocket.Sendln(IntToStr(outtable.recordsleft));
        outtable.fetch_removes:=true;
        while outtable.recordsleft>0 do
          begin
          query:=outtable.fetchrecord;
          ClientSocket.Sendln(query.sql);
          end;
        outtable.Destroy;
        end;
      end;
    query.Destroy;
    end;
  except
    on ex :ENoSuchTable do
      begin
      ClientSocket.Sendln(STDMSG_ERR+ex.Message);
      end;
    on ex :EMyTableError do
      begin
      ClientSocket.Sendln(STDMSG_ERR+ex.Message);
      s:=''; //
      end;
    on ex :EMyQueryError do
      begin
      ClientSocket.Sendln(STDMSG_ERR+ex.Message);
      s:=''; //
      end;
    on ex :EAccessViolation do
      begin
      ClientSocket.Sendln(STDMSG_ERR+EINTERNALERROR_MSG);
      LogMessage(ClientSocket.RemoteHost+':'+ClientSocket.RemotePort+' '+ex.Message, EVENTLOG_WARNING_TYPE, INFO, ID_IERR);
      s:=''; //
      end;
    on ex :Exception do // tu trafia niespodziewane bledy
      begin
      ClientSocket.Sendln(STDMSG_ERR+EINTERNALERROR_MSG);
      LogMessage(ClientSocket.RemoteHost+':'+ClientSocket.RemotePort+' '+ex.Message, EVENTLOG_WARNING_TYPE, INFO, ID_UERR);
      s:=''; //
      end;
  end;
until s='';
ClientSocket.Sendln(STDMSG_GOODBYE);
end;

procedure TDelphiapp2.ServiceExecute(Sender: TService);
begin
  tcpserver1.Active := true;
  while not Terminated do begin
    ServiceThread.ProcessRequests(false);
  end;
  tcpserver1.Active := false;
end;

// zaladowanie biblioteki z obsluga rejestru
// zaladownie tabeli okreslonej w rejestrze
// dodanie wiadomosci w dzienniku systemowym
// w przypadku bledu ladowania pliku dodanie wiadomosci bledu w dzienniku
// zapisanie daty wystartowania serwisu
// odczyt z rejestru i ustawienie numeru portu
// odczyt z rejestru i ustawienie echo_queries
// odczyt z rejestru i ustawienie log_queries i logqueryfile
procedure TDelphiapp2.ServiceStart(Sender: TService; var Started: Boolean);
var buf :string;
begin
try
  buf := ConfigGetValue('tables');
  while (pos(',', buf)>0) and (length(buf)>1) do
    begin
    setlength(tbl, High(tbl)+2);
    tbl[High(tbl)] :=TMyTable.Create;
    LogMessage(TABLELOAD_MSG+copy(buf, 0, pos(',', buf)-1), EVENTLOG_INFORMATION_TYPE, INFO, ID_TBLLOAD);
    tbl[High(tbl)].load(copy(buf, 0, pos(',', buf)-1));
    buf :=copy(buf, pos(',', buf)+1, Length(buf));
    end;

  buf := ConfigGetValue('port');
  tcpServer1.LocalPort:=buf;
  buf := ConfigGetValue('echo_queries');
  if buf='0' then echo_queries:=false
  else echo_queries:=true;
  buf := ConfigGetValue('log_queries');
  if buf='0' then log_queries:=false
  else begin
    log_queries:=true;
    logqueryfile:=buf;
    end;
except
  on ex :Exception do
    begin
    LogMessage(ex.Message, EVENTLOG_ERROR_TYPE, INFO, ID_TBL);
    DoStop;
    end;
end;
LogMessage(LOG_SERVICE_UP, EVENTLOG_INFORMATION_TYPE, INFO, ID_UP);
starttime:=now;
end;

// zapisanie tabeli okreslonej w rejestrze
// dodanie wiadomosci w dzienniku systemowym
// w przypadku bledu zapisywania pliku dodanie wiadomosci bledu w dzienniku
procedure TDelphiapp2.ServiceStop(Sender: TService; var Stopped: Boolean);
var i :integer;
begin
try
for i:=0 to High(tbl) do
  tbl[i].save;
except
  on ex :Exception do
    LogMessage(ex.Message, EVENTLOG_ERROR_TYPE, INFO, ID_TBL);
end;
LogMessage(LOG_SERVICE_DOWN, EVENTLOG_INFORMATION_TYPE, INFO, ID_DOWN);
end;

// laduje biblioteki DLL
procedure TDelphiapp2.ServiceCreate(Sender: TObject);
begin
try
id := LoadLibrary('regopts.dll');
if id=0 then
  Raise EStartupError.Create(ESTARTUPERROR_MSG+' DLL LOAD')
else
  begin
  @ConfigGetValue:=GetProcAddress(id, 'ConfigGetValue');
  if @ConfigGetValue=nil then
    Raise EStartupError.Create(ESTARTUPERROR_MSG+' ConfigGetValue LOAD');
  end;
except
  on ex: Exception do
    begin
    LogMessage('ServiceCreate: '+ex.Message, EVENTLOG_ERROR_TYPE, INFO, ID_CREATE);
    Raise;
    end;
end;
end;

procedure TDelphiapp2.ServiceDestroy(Sender: TObject);
begin
FreeLibrary(id);
end;

end.
